theory Submission
  imports Defs
begin

text \<open>Task 1 (1/10 points)\<close>
theorem G_L: "G \<subseteq> L"
proof 
  fix w assume "w \<in> G"
  thus "w \<in> L"
    by induction (auto simp: L_def)
qed


primrec h :: "_ \<Rightarrow> int" where
  "h [] = 0"
| "h (x # w) = (if x = b then h w + 1 else if x = a then h w - 1 else h w)"

lemma h_altdef: "h w = int (count w b) - int (count w a)"
  by (induction w) auto  

lemma h_append [simp]: "h (w @ w') = h w + h w'"
  by (induction w arbitrary: w') auto

lemma L_altdef: "L = {w. set w \<subseteq> {a, b} \<and> h w = 0}"
  by (auto simp: L_def h_altdef)

text \<open>
  "Intermediate value theorem" for integer sequences: if two adjacent values in the sequence \<open>f\<close>
  differ by at most 1, then any integer between \<open>f 0\<close> and \<open>f n\<close> must occur between \<open>f 0\<close> and \<open>f n\<close>.
\<close>
lemma int_IVT:
  fixes f :: "nat \<Rightarrow> int"
  assumes "\<And>k. k < n \<Longrightarrow> \<bar>f (Suc k) - f k\<bar> \<le> 1"
  assumes "c \<in> {min (f 0) (f n).. max (f 0) (f n)}"
  shows   "\<exists>k. k \<le> n \<and> f k = c"
  using assms
proof (induction n arbitrary: f)
  case 0
  thus ?case by auto
next
  case (Suc n)
  show ?case
  proof (cases "f 0 = c")
    case True
    thus ?thesis by auto
  next
    case False
    from Suc(2)[of 0] have *: "f (Suc 0) = f 0 \<or> f (Suc 0) = f 0 + 1 \<or> f (Suc 0) = f 0 - 1"
      by auto
    have c: "c \<in> {min (f 1) (f (Suc n)).. max (f 1) (f (Suc n))}"
      using False * Suc(2,3) by (auto simp: min_def max_def split: if_splits)
    have "\<exists>k\<le>n. f (k + 1) = c"
      using c Suc(2) by (intro Suc.IH) auto
    thus ?thesis by auto
  qed
qed

definition h' where "h' w k = h (take k w)"

lemma h'_diff: "\<bar>h' w (Suc k) - h' w k\<bar> \<le> 1"
  by (induction w arbitrary: k) (auto simp: h'_def take_Cons split: nat.splits)

lemma aux1:
  assumes h_w: "h w = 2"
  shows   "\<exists>x y. w = x @ y \<and> h x = 1 \<and> h y = 1"
proof -
  have "\<exists>k. k \<le> length w \<and> h' w k = 1"
    by (intro int_IVT h'_diff) (auto simp: h'_def h_w)
  then obtain k where k: "k \<le> length w" "h' w k = 1"
    by auto
  define x y where "x = take k w" and "y = drop k w"
  have w_eq: "w = x @ y" by (simp add: x_def y_def)
  moreover have h_x: "h x = 1"
    using k by (simp add: h'_def x_def)
  moreover have "h y = 1"
    using h_w by (auto simp: w_eq h_x)
  ultimately show ?thesis by blast
qed

lemma aux2:
  assumes h_w: "h w = -2"
  shows   "\<exists>x y. w = x @ y \<and> h x = -1 \<and> h y = -1"
proof -
  have "\<exists>k. k \<le> length w \<and> h' w k = -1"
    by (intro int_IVT h'_diff) (auto simp: h'_def h_w)
  then obtain k where k: "k \<le> length w" "h' w k = -1"
    by auto
  define x y where "x = take k w" and "y = drop k w"
  have w_eq: "w = x @ y" by (simp add: x_def y_def)
  moreover have h_x: "h x = -1"
    using k by (simp add: h'_def x_def)
  moreover have "h y = -1"
    using h_w by (auto simp: w_eq h_x)
  ultimately show ?thesis by blast
qed
  
theorem L_G: "L \<subseteq> G"
proof
  fix w assume "w \<in> L"
  then show "w \<in> G" proof (induction w rule: length_induct)
    case (1 w)
    hence IH: "w' \<in> G" if "w' \<in> L" "length w' < length w" for w'
      using 1 that by auto
    from \<open>w \<in> L\<close> have "set w \<subseteq> {a,b}"
      unfolding L_def by auto
    
    show ?case proof (cases "w = []")
      case True
      thus ?thesis
        by (auto intro: G.intros)
    next
      case False
      with \<open>set w \<subseteq> _\<close> have CH: "hd w \<in> {a, b}"
        by (cases w) auto
      from False \<open>set w \<subseteq> _\<close>
        have CT: "last w \<in> {a, b}" by (cases w rule: rev_cases) auto

      obtain xh xl u where [simp]: "w = xh # u @ [xl]"
      proof (cases w)
        case (Cons xh w')
        thus ?thesis using that
          using \<open>w \<in> L\<close> by (cases w' rule: rev_cases) (auto simp: L_altdef)
      qed (use False in auto)
        
      from CH CT have "xh\<in>{a,b}" "xl\<in>{a,b}" by auto  
      thus ?thesis
      proof safe
        assume [simp]: "xh=a" "xl=a"
        have "\<exists>x y. u = x @ y \<and> h x = 1 \<and> h y = 1"
          using "1.prems" by (intro aux1) (auto simp: L_def h_altdef)
        then obtain x y where xy: "u = x @ y" "h x = 1" "h y = 1"
          by blast
        have "(a # x) @ (y @ [a]) \<in> G"
          by (rule G.intros; rule IH)
             (use xy \<open>set w \<subseteq> _\<close> in \<open>auto simp: L_altdef algebra_simps\<close>)
        thus ?thesis using xy by simp
      next
        assume [simp]: "xh=b" "xl=b"
        have "\<exists>x y. u = x @ y \<and> h x = -1 \<and> h y = -1"
          using "1.prems" by (intro aux2) (auto simp: L_def h_altdef)
        then obtain x y where xy: "u = x @ y" "h x = -1" "h y = -1"
          by blast
        have "(b # x) @ (y @ [b]) \<in> G"
          by (rule G.intros; rule IH)
             (use xy \<open>set w \<subseteq> _\<close> in \<open>auto simp: L_altdef algebra_simps\<close>)
        thus ?thesis using xy by simp
      next
        assume [simp]: "xh = a" "xl = b"
        from \<open>w \<in> L\<close> have "u \<in> G"
          by (intro IH) (auto simp: L_altdef)
        thus ?thesis by (auto intro: G.intros)
      next
        assume [simp]: "xh = b" "xl = a"
        from \<open>w \<in> L\<close> have "u \<in> G"
          by (intro IH) (auto simp: L_altdef)
        thus ?thesis by (auto intro: G.intros)
      qed
    qed      
  qed
qed

end
