theory Submission
imports Defs
begin

(***
It is well known that many functions operating on lists can be rewritten in terms of fold (right).
For instance, length = fold (\<lambda> x n \<rightarrow> n + 1) 0 and map f = fold (\<lambda> x xs \<rightarrow> f x :: xs) [].
Can you prove it ?
***)

lemma length_is_fold:
  "length xs = foldr (\<lambda>x n. n +1) xs 0"
  by (induct xs) auto

lemma map_is_fold:
  "map f xs = foldr (\<lambda>x ys. f x # ys) xs []"
  by (induct xs) auto

(***
But how strong fold actually is ?
It turns out that it is actually the unique solution to this sort of recursive definition:
  g []      = v             <-> g = fold f v
  g (x::xs) = f x (g xs)
Prove it.
***)


theorem universal_property_fold:
  fixes g :: "'a list \<Rightarrow> 'b" and f :: "'a \<Rightarrow> 'b \<Rightarrow> 'b"
  shows "(\<forall>x xs. g (x#xs) = f x (g xs)) \<longleftrightarrow> (\<forall>l. g l = foldr f l (g []))"
  apply rule
  apply clarsimp
  subgoal for l
    apply (induction l)
    apply auto 
    done
  subgoal premises P
    apply clarsimp
  proof -
    fix x xs 
    note P
    show "g (x # xs) = f x (g xs)"
      apply (subst (2) P)
      apply (subst (1) P)
      apply auto
      done
  qed
  done
    

lemma universal_property_foldI: 
    "(\<forall>x xs. g (x#xs) = f x (g xs)) \<Longrightarrow> g l = foldr f l (g [])"
  unfolding universal_property_fold by blast

lemma universal_property_foldI2: 
    "(\<forall>x xs. g (x#xs) = f x (g xs)) \<Longrightarrow> (g []) = n \<Longrightarrow> g l = foldr f l n"
  unfolding universal_property_fold by blast

(***
Using the universal property, prove that filter and fold_left can be rewritten as a fold.
***)

lemma filter_with_fold:
  fixes p :: "'a \<Rightarrow> bool"
  shows            "\<exists>a g. (\<forall>xs. filter p xs = foldr g xs a)"
  apply (rule exI[where x="[]"])
  apply (rule exI[where x="\<lambda>x xs. if p x then x#xs else xs"])
  apply clarify
  apply (rule universal_property_foldI2)
  apply auto
  done

lemma aux: "(\<lambda>b. foldl f b xs) = foldr (\<lambda>g F s. F (f s g)) xs id"  
  apply (rule universal_property_foldI2)
  apply auto
  done
  
lemma foldl_with_fold:
  fixes f :: "'a \<Rightarrow> 'b \<Rightarrow> 'a"
  shows            
  "\<exists>a g. (\<forall>b xs. foldl f b xs = foldr g xs a b)"
  apply (rule exI[where x="id"])
  apply (rule exI[where x="\<lambda>x F s. F (f s x)"])
  apply clarify
  using aux
  by metis  

lemma fib_list_tail'_is_fold:
  "\<exists>f c. fib_list_tail' xs = foldr f xs c"
  apply (rule exI[where x="\<lambda>_ F a b. F (a+b) a"])
  apply (rule exI[where x="\<lambda>a b. b"])
  apply (rule universal_property_foldI2)
  apply auto
  done

lemma fib_list_tail'_add [simp]:
   "fib_list_tail' n a b + fib_list_tail' n a' b' = fib_list_tail' n (a + a') (b + b')"
  by (induction n arbitrary: a a' b b') (auto simp: algebra_simps)

lemma fib_list_is_fib_list_tail_aux: "fib_list (n @ m) = fib_list_tail' n (fib_list (() # m)) (fib_list m)"
  by (induction n arbitrary: m rule: fib_induct.induct) (auto simp: algebra_simps)

lemma fib_list_is_fib_list_tail:
  "fib_list n = fib_list_tail n"
  unfolding fib_list_tail_def using fib_list_is_fib_list_tail_aux[of n "[]"]
  by simp 
  
theorem fib_list_is_fold:
  "\<exists>f a. fib_list n = foldr f n a (Suc 0) (Suc 0)"
  by (metis One_nat_def fib_list_is_fib_list_tail fib_list_tail'_is_fold fib_list_tail_def)

end
