theory Submission
imports Defs
begin

lemma has_global_maximum:
  fixes terrain :: "int \<Rightarrow> int \<Rightarrow> nat"
    and (* delimiters *) lx ux ly uy :: int
    and (* one element *) ix iy :: int
  assumes  w: "within lx ux ly uy ix iy"
    and distinct: "\<And>ax ay bx by. (ax,ay) \<noteq> (bx,by) \<Longrightarrow> terrain ax ay \<noteq> terrain bx by"
    and border: "\<And>bx by. onborder lx ux ly uy bx by \<Longrightarrow> terrain bx by < terrain ix iy" 
  shows "\<exists>tx ty. within lx ux ly uy tx ty \<and>
           (\<forall>d\<in>{(0,1),(1,0),(-1,0),(0,-1)}. terrain tx ty > terrain (tx+fst d) (ty+snd d))"
proof -
  let ?t = "(\<lambda>(x,y). terrain x y)"
  let ?D = "{lx<..<ux} \<times> {ly<..<uy}"
  from w have ixy: "(ix, iy) \<in> ?D" by (auto simp: within_def)

  define M where "M = Max (?t ` ?D)"
  have M_ge: "terrain x y \<le> M" if "within lx ux ly uy x y" for x y
  proof -
    have "?t (x, y) \<in> ?t ` ?D"
      using that by (intro imageI) (auto simp: within_def)
    hence "?t (x, y) \<le> M" unfolding M_def
      by (intro Max_ge) auto
    thus ?thesis by simp
  qed

  have "M \<in> ?t ` ?D"
    unfolding M_def using ixy by (intro Max_in) auto
  then obtain tx ty where [simp]: "M = ?t (tx, ty)" and "(tx, ty) \<in> ?D"
    by force
  hence txy: "within lx ux ly uy tx ty" by (auto simp: within_def)

  from M_ge have txy_ge: "terrain tx ty \<ge> terrain x y" if "within lx ux ly uy x y" for x y
    using M_ge[of x y] that by auto

  show ?thesis
  proof (rule exI[of _ tx], rule exI[of _ ty], intro conjI ballI)
    fix d :: "int \<times> int" assume d: "d \<in> {(0, 1), (1, 0), (- 1, 0), (0, - 1)}"
    show "terrain (tx + fst d) (ty + snd d) < terrain tx ty"
    proof (cases "within lx ux ly uy (tx + fst d) (ty + snd d)")
      case True
      hence "terrain (tx + fst d) (ty + snd d) \<le> terrain tx ty"
        by (intro txy_ge)
      moreover have "terrain (tx + fst d) (ty + snd d) \<noteq> terrain tx ty"
        using d by (intro distinct) auto
      ultimately show ?thesis by simp
    next
      case False
      with txy d have "onborder lx ux ly uy (tx + fst d) (ty + snd d)"
        by (fastforce simp: onborder_def within_def)
      with border have "terrain (tx + fst d) (ty + snd d) < terrain ix iy" by blast
      also have "\<dots> \<le> terrain tx ty"
        using w by (intro txy_ge) auto
      finally show ?thesis .
    qed
  qed fact+
qed

end
